#include <math.h>
#include "wcolgradgen.h"

// versions one and two were stripped

/* ------ BEGIN version 3 ------ */
/* BEGIN - polynomial creation, setting and deletion */

/* recreates the lookup table for a polynomial if necessary */
int make_plut(ppolynomial PN)
{
  int i;
  int j;
  float cx,cy;

  if(!PN) return -1; /* error, not a valid polynomial */
  if(!PN->changed) return 0; /* nothing to do, because polynomial has not changed
                                since last computation */

  if(PN->plut) free(PN->plut); /* kill old table */

  /* create new table */
  PN->plut=NULL;
  PN->plut=(int *)malloc(PN->w * sizeof(int));
  if(!PN->plut) return -1; /* error not enough memory */

  for(i=0; i<PN->w; i++) /* compute all function values along the entire width */
  {
    cx = (-(PN->x - i)) / (float)PN->unit;

    /* naive method of computing the function value */
    /* cy=0; */
    /* for(j=(int)PN->he; j>1; j--) /* compute function value down to co[2]*x^2 */
    /* { */
    /*   cy += PN->co[j] * pow(cx,(float)j); 
    /* } */
    /* cy += (PN->co[1] * cx) + PN->co[0]; /* add the x^1 and constant ((x^0)==1) */
    
    /* using "Horner Scheme" to compute function value */
    cy=PN->co[PN->he]*cx;
    for(j=(int)PN->he-1; j>0; j--)
      cy = (PN->co[j] + cy)*cx;
    cy = cy + PN->co[0];

    PN->plut[i]=PN->y+(int)(cy*PN->unit); /* finally add the value to the table */
  }
  PN->changed=0;
  return 0; /* all fine, new table is good */
}

/* returns a pointer to a valid polynomial, or null on failure */
/* last parameters are the he+1 coefficients in same order as described in PN_DESC*/
ppolynomial make_polynomial(unsigned int he, int unit, int x, int y, ...)
{
  va_list marker; /* variable argument list <stdarg.h> */
  int i;
  ppolynomial work=NULL;
  work = (ppolynomial)malloc(sizeof(polynomial));
  if (!work) return NULL; /* error not enough memory */
  
  (work)->he=he;
  (work)->co=NULL;

  (work)->co=(double *)malloc((he+1) * sizeof(double));
  if(!((work)->co)) /* error not enough memory */
  {
    free(work);
    return NULL;
  } else /* init coefficients */
  {
    va_start(marker, y); /* init variable argument list */
    for(i=(int)he; i>=0; i--)
      (work)->co[i]=va_arg(marker, double); /* evaluate variable argument list */
  }
  va_end(marker); /* kill variable argument list */

  (work)->unit=(unit!=0)? unit : 1; /* set safe unit to prevent div by 0 */
  (work)->x=x;
  (work)->y=y;
  (work)->w=0; /* will be set by render...3 function */
  (work)->plut=NULL; /* lookup table will be generated in "make_plut" on the first time render */
  (work)->changed=1;
  
  return work;
}

/* adjusting the polynome after createn, each will cause the lookup table to
   be recalculated on the first render */
void set_pn_co(ppolynomial PN, unsigned int which, float val)
{
  if(PN)
    if((which)<=(PN->he))
      if(PN->co)
        PN->co[which]=val;
}
void set_pn_unit(ppolynomial PN, int unit)
{
  if(PN)
  {
    PN->unit=(unit!=0)? unit : 1; /* set safe unit to prevent div by 0 */
    PN->changed=1;
  }
}
void set_pn_x(ppolynomial PN,int x)
{
  if(PN)
  {
    PN->x=x;   
    PN->changed=1;
  }
}
void set_pn_y(ppolynomial PN,int y)
{
  if(PN)
  {
    PN->y=y;
    PN->changed=1;
  }
}
void set_pn_w(ppolynomial PN,int w)
{
  if(PN)
  {
    PN->w=w;
    PN->changed=1;
  }
}

/* deletion of a polynome, will be used by "kill_ca_list3" if necessary, so
   inside this programme the user won't need to call this */
void kill_polynomial(ppolynomial PN)
{
  if(!PN) return;
  else
  {
    if(PN->co) { free(PN->co); PN->co=NULL; }
    if(PN->plut) { free(PN->plut); PN->plut=NULL; }
  }
}

/* END - polynomial creation, setting and deletion */

int init_ca_list3(col_att_list3* pcal)
{
  (pcal)->link_count=0;
  (pcal)->plist=NULL;
  return 0;
}

int kill_ca_list3(col_att_list3* pcal)
{
  pcol_att3 current=pcal->plist;
  pcol_att3 next=NULL;

  while(current)
  {
    next=current->pnext;
    if(current->PN) { kill_polynomial(current->PN); current->PN=NULL; }
    free(current);
    current=next;
  }
  (pcal)->link_count=0;
  (pcal)->plist=NULL;
  return 0;
}

int add_to_ca_list3(col_att_list3* pcal, col_att_type t, col_att_mode m, float p, float exp, int x, int y, int r, int g, int b, int range, ppolynomial PN)
{
  pcol_att3 first=pcal->plist;
  pcol_att3 work=NULL;
  work=(CA_DESC3 *)malloc(sizeof(col_att3));
  if(work)
  {
    (work)->t=t;
    (work)->m=m;
    (work)->p=p;
    (work)->exp=exp;
    (work)->x=x;
    if (t==CA_GLOBAL) { (work)->x=100-x; if( ((work)->x < 0) || ((work)->x > 100) ) (work)->x=100; }
    (work)->y=y;
    (work)->r=r;
    (work)->g=g;
    (work)->b=b;
    (work)->range=abs(range);
    if (t==CA_GLOBAL) (work)->range=100; /* that way x will be interpreted in % of color to mix in */
    (work)->PN=PN;
    (work)->pnext=first;
    (pcal)->link_count+=1;
    (pcal)->plist=work;
    return 0;
  }
  else return -1; /* not enough memory */
}

/* computes for a given pixel(x,y) the distance to the
   specified "color attractor" (version 3) */
float distance_to_ca3(int x, int y, pcol_att3 ca)
{
  if (ca->t==CA_GLOBAL)
    return abs(ca->x);

  if (ca->t==CA_POINT)
  {
    float vx=ca->x-x;
    float vy=ca->y-y;
    return sqrt((vx*vx)+(vy*vy));
  }

  if (ca->t==CA_HBAR)
    return abs(ca->y-y);

  if (ca->t==CA_VBAR)
    return abs(ca->x-x);

  if (ca->t==CA_POLYNOMIAL)
  {
    /* uses lookup table for calculation if available else returns 0
       to fall back to weird but at least uncrashing behaviour */
    if(ca->PN)
      if(ca->PN->plut)
        return abs((ca->PN->plut[x])-y);
      else return 0;
  }

  return 0.0; /* invalid ca type */
}

void render_col_grad_to_bitmap3(BITMAP *bmp, col_att_list3* pcal)
{
  if (pcal==NULL) return; /* nothing to do */
  else if (bmp==NULL) return; /* nothing to do */
  else
  {
    int i;
    int x,y;
    int w=bmp->w;
    int h=bmp->h;
    int r,g,b;
    float fr,fg,fb;
    pcol_att3 next=NULL;
    float d;
    float calc; // used to prevent unnecessary multiple calculations of the same value


    /* recalculate polynomial lookup tables */
    next=pcal->plist;
    for(i=0; i<pcal->link_count; i++)
    {
      if(next->PN)
        if( (next->PN->w) != bmp->w) /* if target bitmap has different width than lookup table... */
          set_pn_w(next->PN,bmp->w); /* makes sure that the lookup table will
                                        be calculated for evey x-pixel position */
      make_plut(next->PN);
      next=next->pnext;
    }

    for(y=0; y<h; y++) for(x=0; x<w; x++) /* for every pixel...*/
    {
      r=0; g=0; b=0;
      fr=0; fg=0; fb=0;
      d=0;

      /* ...compute the color components for the pixel */
      next=pcal->plist;
      
      for(i=0; i<pcal->link_count; i++)
      {
        d = distance_to_ca3(x,y,next);
        
        if((d <= next->range)&&(next->range!=0))
        {
          calc = (1.0 - pow(d / next->range,next->exp)) * next->p;

          switch(next->m) /* evaluate CA_MODE */
          {
            case CA_ATTRACT:
              fr += calc * next->r;
              fg += calc * next->g;
              fb += calc * next->b;
            break;
            case CA_ABSORB:
              fr -= calc * next->r;
              fg -= calc * next->g;
              fb -= calc * next->b;
            break;
          }
        }
        next=next->pnext;
      }
      r=(int)fr;
      g=(int)fg;
      b=(int)fb;


      /* prevent invalid color values */
      if(r<0)
        r=0;
        else if(r>255)
          r=255;

      if(g<0)
        g=0;
        else if(g>255)
          g=255;

      if(b<0)
        b=0;
        else if(b>255)
          b=255;

      putpixel(bmp,x,y,makecol(r,g,b));
      if(key[KEY_ESC]) break; /* if accidantally called on "screen", this is an emergency exit */
    }
  }
}
/* ------ END version 3 ------ */